﻿// Simple class to send remote control messages to Sony TVs (and other supported devices).
//
// Sony TVs have a webserver running, which responds to IRCC commands and executes them.
// Clients who want to control the TV need to be registered first. 
// TV will also send a full list of supported commands when making a request to "/cers/api/getRemoteCommandList"
//
// Usage:
//
// 1. Create instance of the class with TV IP-address (and port if you want, otherwise default 80 is used).
//   
// 2. Register client with RegisterClient() method (only needed once). If this is the first time this client (= mac address) is registering,
//    TV will show a message to user and user needs to click OK. If client is already registered TV will just respond with OK immediately.
//
// 3. Get supported commands list with GetCommandList().
//
// 4. Send commands to TV with SendCommand(). For example SendCommand(commands["Mode3D"]).     
//
//
// When this class is instantiated, it will add commands needed to toggle 3D mode to commandlist by default.
// So to toggle 3D mode with registered client, create instance and just call Enable3D or Disable3D.
// To change back to 2D mode, Disable3D() method needs to know which 3D mode was selected (parameter mode).
// 3D modes work at least with HX series TV's, but others (NX, EX, etc.) might work too.
// 
//
// This class is based on this project: https://github.com/mlampard/ruby_cers_remote ,
// with following CopyRight information: © Luke Mcildoon 2012 - MIT licenced Additions © 2012-2013 Mike Lampard ,
// and license information: http://opensource.org/licenses/MIT
//


using System;
using System.Collections.Specialized;
using System.Diagnostics;
using System.Xml;
using System.Text;
using System.Net;
using System.Threading;
using System.IO;

namespace SonyIpRemote
{
  class SonyIpRemote
  {

    public enum Mode3D
    {
      SideBySide = 0,
      OverUnder = 1
    }

    private readonly string DeviceID = "MediaPortal"; // This name will show up on TV's registered controllers list.

    private readonly string PostCommandUrl = "/IRCC";
    private readonly string RegUrlStart = "/cers/api/register?name=";
    private readonly string RegUrlEnd = "&registrationType=initial&deviceId=MediaRemote";
    private readonly string RegisterDeviceUrl = "/cers/api/register?name=DeviceName&registrationType=initial&deviceId=MediaRemote";
    private readonly string ListCommandsUrl = "/cers/api/getRemoteCommandList";
    
    private string TVIP;

    public StringDictionary commands; 

    // TV IP in format "192.168.0.100", port is 80 by default.
    public SonyIpRemote(string IP, ushort port)
    {
      commands = new StringDictionary();

      commands.Add("Mode3D", "AAAAAgAAAHcAAABNAw==");
      commands.Add("Confirm", "AAAAAQAAAAEAAABlAw==");
      commands.Add("Down", "AAAAAQAAAAEAAAB1Aw==");
      commands.Add("Return", "AAAAAgAAAJcAAAAjAw==");

      RegisterDeviceUrl = RegUrlStart + DeviceID + RegUrlEnd;

      TVIP = IP + ":" + port.ToString();
    }

    public SonyIpRemote(string IP) : this(IP, 80) { }

    // This will register the client with TV. Only one successful call/device is needed, but if already registered client
    // tries to register, TV will just respond OK.

    public bool RegisterClient()
    {
      HttpWebRequest request = (HttpWebRequest)WebRequest.Create("http://" + TVIP + RegisterDeviceUrl);

      request.Headers.Add("X-CERS-DEVICE-INFO", DeviceID);
      request.KeepAlive = false;
      request.Proxy = null;
      request.Timeout = 30000;  // There is a 30 second countdown running on the TV side if client is registering for the first time.
                                // User needs to click OK and confirm before remote commands are accepted.

      try
      {
        using (WebResponse response = (HttpWebResponse)request.GetResponse()) { }
      }
      catch // TV will respond with 40X code if registering fails
      {
        Debug.Write("Registering client failed.");
        return false;
      }

      return true;
    }

    // Fetches command list from TV and adds all commands to dictionary 'commands'.

    public void GetCommandList()
    {
      HttpWebRequest request = (HttpWebRequest)WebRequest.Create("http://" + TVIP + ListCommandsUrl);

      request.Headers.Add("X-CERS-DEVICE-INFO", DeviceID);
      request.Headers.Add("X-CERS-DEVICE-ID", "MediaRemote");
      request.KeepAlive = false;
      request.Proxy = null;

      string result;
      try
      {
        using (WebResponse response = (HttpWebResponse)request.GetResponse())
        {
          using (Stream responseStream = response.GetResponseStream())
          {
            using (StreamReader sr = new StreamReader(responseStream))
            {
              result = sr.ReadToEnd();
            }
          }
        }

        XmlDocument xml = new XmlDocument();
        xml.LoadXml(result);

        XmlNodeList list = xml.GetElementsByTagName("command");

        commands.Clear();

        for (int i = 0; i < list.Count; i++)
        {
          string name = list[i].Attributes["name"].Value;
          string value = list[i].Attributes["value"].Value;

          Debug.WriteLine("Command: " + name + " Value: " + value);

          commands.Add(name, value);
        }
      }
      catch
      {
        Debug.WriteLine("Fetching command list failed.");
      }

    }


    private string Envelope(string command)
    {
      StringBuilder msg = new StringBuilder();
      msg.AppendLine("<?xml version=\"1.0\"?>");
      msg.AppendLine("<s:Envelope xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\" s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\">");
      msg.AppendLine("<s:Body>");
      msg.AppendLine("<u:X_SendIRCC xmlns:u=\"urn:schemas-sony-com:service:IRCC:1\">");
      msg.AppendFormat("<IRCCCode>{0}</IRCCCode>", command);
      msg.AppendLine();
      msg.AppendLine("</u:X_SendIRCC>");
      msg.AppendLine("</s:Body>");
      msg.AppendLine("</s:Envelope>");

      return msg.ToString();
    }

    // Send command to tv. Parameter 'command' in IRCC format, e.g. 'AAAAAQAAAAEAAABlAw==' .

    public string sendCommand(string command)
    {
      HttpWebRequest request = (HttpWebRequest)WebRequest.Create("http://" + TVIP + PostCommandUrl);

      request.Method = "POST";
      request.Headers.Add("X-CERS-DEVICE-INFO", DeviceID);
      request.Headers.Add("X-CERS-DEVICE-ID", "MediaRemote");
      request.Headers.Add("SOAPAction", "urn:schemas-sony-com:service:IRCC:1#X_SendIRCC");
      request.ContentType = "text/xml; charset=UTF-8";
      request.KeepAlive = false;
      request.Proxy = null;
      request.Timeout = 5000;
      
      byte[] byteArray = Encoding.UTF8.GetBytes(Envelope(command));

      request.ContentLength = byteArray.Length;

      Stream dataStream = request.GetRequestStream();
      dataStream.Write(byteArray, 0, byteArray.Length);
      dataStream.Close();

      string result;

      try
      {

        using (WebResponse response = (HttpWebResponse)request.GetResponse())
        {
          using (Stream responseStream = response.GetResponseStream())
          {
            using (StreamReader sr = new StreamReader(responseStream))
            {
              result = sr.ReadToEnd();
            }
          }
        }
      }
      catch
      {
        Debug.WriteLine("Sending command failed");
        result = "ERROR";
      }

      return result;
    }

    public void Enable3D(Mode3D mode = Mode3D.SideBySide)
    {
      sendCommand(commands["Mode3D"]);

      Thread.Sleep(500); // This is necessary, the 3D-menu takes time to popup

      sendCommand(commands["Confirm"]);
      sendCommand(commands["Down"]);

      if (mode == Mode3D.SideBySide)
      {
        sendCommand(commands["Down"]);
      }

      sendCommand(commands["Confirm"]);

      Thread.Sleep(1000); // While the 3D mode is switching TV doesn't respond for a while

      sendCommand(commands["Return"]);

    }

    public void Disable3D(Mode3D mode = Mode3D.SideBySide)
    {
      sendCommand(commands["Mode3D"]);

      Thread.Sleep(500); // This is necessary, the 3D-menu takes time to popup

      sendCommand(commands["Confirm"]);
      sendCommand(commands["Down"]);
      sendCommand(commands["Down"]);

      if (mode == Mode3D.OverUnder)
      {
        sendCommand(commands["Down"]);
      }

      sendCommand(commands["Confirm"]);

      Thread.Sleep(1000); // While the 3D mode is switching TV doesn't respond for a while

      sendCommand(commands["Return"]);
    }
  }
}




/*

HX-series default commands: 
 
<?xml version="1.0"?>
<remoteCommandList>
    <command name="Confirm" type="ircc" value="AAAAAQAAAAEAAABlAw==" />
    <command name="Up" type="ircc" value="AAAAAQAAAAEAAAB0Aw==" />
    <command name="Down" type="ircc" value="AAAAAQAAAAEAAAB1Aw==" />
    <command name="Right" type="ircc" value="AAAAAQAAAAEAAAAzAw==" />
    <command name="Left" type="ircc" value="AAAAAQAAAAEAAAA0Aw==" />
    <command name="Home" type="ircc" value="AAAAAQAAAAEAAABgAw==" />
    <command name="Options" type="ircc" value="AAAAAgAAAJcAAAA2Aw==" />
    <command name="Return" type="ircc" value="AAAAAgAAAJcAAAAjAw==" />
    <command name="Num1" type="ircc" value="AAAAAQAAAAEAAAAAAw==" />
    <command name="Num2" type="ircc" value="AAAAAQAAAAEAAAABAw==" />
    <command name="Num3" type="ircc" value="AAAAAQAAAAEAAAACAw==" />
    <command name="Num4" type="ircc" value="AAAAAQAAAAEAAAADAw==" />
    <command name="Num5" type="ircc" value="AAAAAQAAAAEAAAAEAw==" />
    <command name="Num6" type="ircc" value="AAAAAQAAAAEAAAAFAw==" />
    <command name="Num7" type="ircc" value="AAAAAQAAAAEAAAAGAw==" />
    <command name="Num8" type="ircc" value="AAAAAQAAAAEAAAAHAw==" />
    <command name="Num9" type="ircc" value="AAAAAQAAAAEAAAAIAw==" />
    <command name="Num0" type="ircc" value="AAAAAQAAAAEAAAAJAw==" />
    <command name="Num11" type="ircc" value="AAAAAQAAAAEAAAAKAw==" />
    <command name="Num12" type="ircc" value="AAAAAQAAAAEAAAALAw==" />
    <command name="Power" type="ircc" value="AAAAAQAAAAEAAAAVAw==" />
    <command name="Display" type="ircc" value="AAAAAQAAAAEAAAA6Aw==" />
    <command name="VolumeUp" type="ircc" value="AAAAAQAAAAEAAAASAw==" />
    <command name="VolumeDown" type="ircc" value="AAAAAQAAAAEAAAATAw==" />
    <command name="Mute" type="ircc" value="AAAAAQAAAAEAAAAUAw==" />
    <command name="Audio" type="ircc" value="AAAAAQAAAAEAAAAXAw==" />
    <command name="SubTitle" type="ircc" value="AAAAAgAAAJcAAAAoAw==" />
    <command name="Yellow" type="ircc" value="AAAAAgAAAJcAAAAnAw==" />
    <command name="Blue" type="ircc" value="AAAAAgAAAJcAAAAkAw==" />
    <command name="Red" type="ircc" value="AAAAAgAAAJcAAAAlAw==" />
    <command name="Green" type="ircc" value="AAAAAgAAAJcAAAAmAw==" />
    <command name="Play" type="ircc" value="AAAAAgAAAJcAAAAaAw==" />
    <command name="Stop" type="ircc" value="AAAAAgAAAJcAAAAYAw==" />
    <command name="Pause" type="ircc" value="AAAAAgAAAJcAAAAZAw==" />
    <command name="Rewind" type="ircc" value="AAAAAgAAAJcAAAAbAw==" />
    <command name="Forward" type="ircc" value="AAAAAgAAAJcAAAAcAw==" />
    <command name="Prev" type="ircc" value="AAAAAgAAAJcAAAA8Aw==" />
    <command name="Next" type="ircc" value="AAAAAgAAAJcAAAA9Aw==" />
    <command name="Replay" type="ircc" value="AAAAAgAAAJcAAAB5Aw==" />
    <command name="Advance" type="ircc" value="AAAAAgAAAJcAAAB4Aw==" />
    <command name="TopMenu" type="ircc" value="AAAAAgAAABoAAABgAw==" />
    <command name="PopUpMenu" type="ircc" value="AAAAAgAAABoAAABhAw==" />
    <command name="Eject" type="ircc" value="AAAAAgAAAJcAAABIAw==" />
    <command name="Rec" type="ircc" value="AAAAAgAAAJcAAAAgAw==" />
    <command name="SyncMenu" type="ircc" value="AAAAAgAAABoAAABYAw==" />
    <command name="ClosedCaption" type="ircc" value="AAAAAgAAAKQAAAAQAw==" />
    <command name="Teletext" type="ircc" value="AAAAAQAAAAEAAAA/Aw==" />
    <command name="ChannelUp" type="ircc" value="AAAAAQAAAAEAAAAQAw==" />
    <command name="ChannelDown" type="ircc" value="AAAAAQAAAAEAAAARAw==" />
    <command name="Input" type="ircc" value="AAAAAQAAAAEAAAAlAw==" />
    <command name="GGuide" type="ircc" value="AAAAAQAAAAEAAAAOAw==" />
    <command name="EPG" type="ircc" value="AAAAAgAAAKQAAABbAw==" />
    <command name="DOT" type="ircc" value="AAAAAgAAAJcAAAAdAw==" />
    <command name="Analog" type="ircc" value="AAAAAgAAAHcAAAANAw==" />
    <command name="Exit" type="ircc" value="AAAAAQAAAAEAAABjAw==" />
    <command name="Digital" type="ircc" value="AAAAAgAAAJcAAAAyAw==" />
    <command name="BS" type="ircc" value="AAAAAgAAAJcAAAAsAw==" />
    <command name="CS" type="ircc" value="AAAAAgAAAJcAAAArAw==" />
    <command name="BSCS" type="ircc" value="AAAAAgAAAJcAAAAQAw==" />
    <command name="Ddata" type="ircc" value="AAAAAgAAAJcAAAAVAw==" />
    <command name="InternetWidgets" type="ircc" value="AAAAAgAAABoAAAB6Aw==" />
    <command name="InternetVideo" type="ircc" value="AAAAAgAAABoAAAB5Aw==" />
    <command name="SceneSelect" type="ircc" value="AAAAAgAAABoAAAB4Aw==" />
    <command name="Mode3D" type="ircc" value="AAAAAgAAAHcAAABNAw==" />
    <command name="iManual" type="ircc" value="AAAAAgAAABoAAAB7Aw==" />
    <command name="Wide" type="ircc" value="AAAAAgAAAKQAAAA9Aw==" />
    <command name="Jump" type="ircc" value="AAAAAQAAAAEAAAA7Aw==" />
    <command name="PAP" type="ircc" value="AAAAAgAAAKQAAAB3Aw==" />
    <command name="MyEPG" type="ircc" value="AAAAAgAAAHcAAABrAw==" />
    <command name="ProgramDescription" type="ircc" value="AAAAAgAAAJcAAAAWAw==" />
    <command name="WriteChapter" type="ircc" value="AAAAAgAAAHcAAABsAw==" />
    <command name="TrackID" type="ircc" value="AAAAAgAAABoAAAB+Aw==" />
    <command name="TenKey" type="ircc" value="AAAAAgAAAJcAAAAMAw==" />
    <command name="AppliCast" type="ircc" value="AAAAAgAAABoAAABvAw==" />
    <command name="acTVila" type="ircc" value="AAAAAgAAABoAAAByAw==" />
    <command name="DeleteVideo" type="ircc" value="AAAAAgAAAHcAAAAfAw==" />
    <command name="EasyStartUp" type="ircc" value="AAAAAgAAAHcAAABqAw==" />
    <command name="OneTouchTimeRec" type="ircc" value="AAAAAgAAABoAAABkAw==" />
    <command name="OneTouchView" type="ircc" value="AAAAAgAAABoAAABlAw==" />
    <command name="OneTouchRec" type="ircc" value="AAAAAgAAABoAAABiAw==" />
    <command name="OneTouchRecStop" type="ircc" value="AAAAAgAAABoAAABjAw==" />
</remoteCommandList>

*/